#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <sched.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "io.h"
#include "signal-event.h"
#include "timeout.h"
#include "misc.h"
#include "list.h"
#include "sock-unix.h"
#include "hal-mgr.h"
#include "hal-mgr-conf.h"
#include "main.h"

#define DEBUG_HALMGR

#define warn(fmt, arg...) fprintf(stderr, "warn: %s: "fmt CRLF, \
					__FUNCTION__, ##arg);
#define info(fmt, arg...) fprintf(stderr, "info: %s: "fmt CRLF, \
					__FUNCTION__, ##arg);

#define error(eval, fmt, arg...)			\
do {							\
	fprintf(stderr, "error: %s: "fmt CRLF,		\
				__FUNCTION__, ##arg);	\
	exit(eval);					\
} while (0)


#ifdef DEBUG_HALMGR
#define debug(fmt, arg...) fprintf(stderr, "debug: %s: "fmt CRLF, \
					__FUNCTION__, ##arg);
#else
#define debug(fmt, arg...) do {} while(0) 
#endif

extern char	*__progname;
struct slist	*pi_srv_options_list;

static int 	terminate;
static struct io_context *io;

static void print_usage()
{
	printf("Usage: %s [-ih] [ -c config]\n"
	" -h		print this help\n"
	" -i		stay interactive do not become a daemon\n"
	" -c config	path to a config file\n", __progname);
}

static void init_clients()
{
	struct pi_srv_options *opts;
	struct slist *iter;
	int ret;

	for (iter = pi_srv_options_list; iter != NULL; iter = iter->next) {
		opts = (struct pi_srv_options *)iter->data;

		if (strlen(opts->socketfile) != 0) {
			ret = hal_mgr_client_unix_new(opts->name, opts->socketfile,
						      opts->timeout, opts->side);
		} else if (strlen(opts->inet_addr) != 0) {
			ret = hal_mgr_client_inet_new(opts->name, opts->inet_addr,
						      opts->timeout, opts->side);
		} else  {
			error(1, "neither unix socket nor inet address specified for %s",
									opts->name);
		}
		
		if (ret == -1)
			error(1, "can't connect to %s", opts->name);
	}
}

static void realtime_enter(int priority)
{
	struct sched_param parameters;
	int res;

	parameters.sched_priority = priority;
	res = sched_setscheduler(0, SCHED_RR, &parameters);
	if (res) {
		info("Failed to enter RT SCHED_RR scheduler: %s", strerror(errno));
		return;
	}

	debug("RT scheduler priority %d", priority);
}

static void term_signal(int signo, siginfo_t *si, void *opaque)
{
	switch (signo) {
	case SIGINT:
	case SIGTERM:
		info("exiting by signal %d", signo);
		terminate = 1;
		io_loop_stop(io);
		break;
	}
}

static void accept_event(int fd, short events, void *opaque)
{
        struct sockaddr_un saddr;
	int client_fd;
        socklen_t slen;

        if (events & EVENT_ERROR) {
		warn("got error on socket");
		return;
        }

	if (events & EVENT_READ) {
		slen = sizeof(saddr);
		client_fd = accept(fd, (struct sockaddr *)&saddr, &slen);

		if (client_fd >= 0) {
			debug("accept a new connection %d", client_fd);
			/* We do believe it is a runtime and not something
			 * else. */
			sock_nonblock(client_fd);
			hal_mgr_client_raw_new("runtime", client_fd, client_fd,
						0, HAL_CLT_RUNTIME);
		} else {
			warn("accept()==%d: %s", fd, strerror(errno));
		}
	}
}

int main(int argc, char **argv)
{
	struct hal_mgr_options options;
	char filepath[PATH_MAX];
	char configfile[PATH_MAX];
	char *ptr;
	int no_detach, ret, opt, sfd;
	int cflag;

	no_detach = 1;
	cflag = 0;

	while ((opt = getopt(argc, argv, "ihc:")) != -1) {
		switch (opt) {
		case 'i':
			no_detach = 1;
			break;
		case 'h':
			print_usage();
			exit(0);
		case 'c':
			xstrncpy(filepath, optarg, sizeof(filepath));
			ptr = canonize_path(filepath);
			xstrncpy(configfile, ptr, sizeof(configfile));
			xfree(ptr);
			cflag = 1;
			break;
		default:
			print_usage();
			exit(-1);
		}
	}

	argv += optind;
	argc -= optind;

	if (!cflag)
		error(1, "config file option not found");

	ret = config_read(configfile, &options);

	if (ret != 0)
		error(1, "can't parse a configuration file");

	sigevent_ignore(SIGPIPE);
	sigevent_add(SIGINT, term_signal, NULL, 1);
	sigevent_add(SIGTERM, term_signal, NULL, 1);
	sigevent_init();

	timeout_init();
	io = io_new();

	hal_mgr_init(io);

//	init_clients();

	hal_mgr_clients_connect();

	config_free();

	realtime_enter(options.priority);

	if (strlen(options.socketfile) != 0) {
		sfd = sock_unix_listen(options.socketfile);
		if (sfd == -1)
			error(1, "can't create socket %s", options.socketfile);

		sock_nonblock(sfd);
		io_add(io, sfd, EVENT_READ, accept_event, NULL);
	}

	info("Run a loop.");

	while (!terminate)
		io_loop_run(io);

	hal_mgr_deinit();

	timeout_deinit();

	io_free(&io);
	sigevent_deinit();
	
	ret = unlink(options.socketfile);
	if (ret != 0)
		info("unlink() %s", strerror(errno));

	info("It is over.");

	return 0;
}

